/*------------------------------------------------------------------------------
 *
 *       Filename:  ar.c
 *
 *    Description:  Activity recognition on streaming sensor data.
 *
 *        Version:  2.0
 *        Created:  05/07/2012 06:46:51 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *        Authors:  Narayanan C Krishnan, ckn@eecs.wsu.edu
 *                  Diane J. Cook, cook@eecs.wsu.edu
 *   Organization:  Washington State University
 * ---------------------------------------------------------------------------*/

#include "ar.h"


/* The command line syntax is ar [configfile].  The data file and program
   parameters are specified in the config file. If no config filename is
   provided then the default file of ar.config is assumed. */
int main(int argc, char *argv[])
{
   char fname[MAXSTR];
   int i;

   printf("AR - Reading the config file...");
   if (argc > 1)
      ReadConfig(argv[1]);
   else
   {
      strcpy(fname, "ar.config");
      ReadConfig(fname);
   }
   printf("done\n\n");


   if (argc > 2)
   {
       printf("Reading user specified data file names ...");
       NumFiles = 0;
       for (i=2; i<argc; i++)
       {
           if (NumFiles == 0)
               filenames = (char **) malloc(sizeof(char *));
           else filenames =
               (char **) realloc(filenames, sizeof(char *) * (NumFiles + 1));
           filenames[NumFiles] = (char *) malloc(sizeof(char ) * MAXSTR);
           strcpy(filenames[NumFiles], argv[i]);
           NumFiles++;
       }
   }
   printf("done\n\n");

   if (OutputLevel > 0)
       PrintParms();

   // read the hierarchy
   if (hierarchical == TRUE)
       ReadTree();
   else DefaultTree();

   if (Mode == TRAIN)
   {
       printf("\nEntering training mode\n");

       if (factwind != NULL)
           free(factwind);
       if (fsenact != NULL)
           free(fsenact);

       // Compute the window probabilities
       if (OutputLevel > 0)
           printf("\nComputing probabilities to determine window size...");
       ComputeProbActWind();
       ComputeProbSenAct();
       if (OutputLevel > 0)
           printf("done\n");

       // Compute mutual information between the sensors
       if (Weight > 1)
       {
           if (OutputLevel > 0)
               printf("\nComputing mutual information between the sensors...");
           ComputeMI();
           if (OutputLevel > 0)
               printf("done\n");
       }

       ReadData();             // Read the data
       mkdir(modelfname, S_IRWXU | S_IRWXG | S_IRWXO);    // creat the directory to store the model files and the final config file
       LearnActivityModels();  // Build and save the models
       SaveParams();           // Save the program parameters to a file
   }
   else // TEST mode
   {
       ReadData();             // Read the data
       TestActivityModels();   // Load the activity models and test the data
   }

   FreeMemory();

   return(1);
}


/* Read each line of the config file and set program parameters accordingly. */
void ReadConfig(char *name)
{
   FILE *fp;
   char *cptr, buffer[50000], *tok;
   int i, j;

   fp = fopen(name, "r");
   if (fp == NULL)
   {
      printf("Unable to open the configuration file: %s\n", name);
      exit(1);
   }

   cptr =  fgets(buffer, 40000, fp);

   while (cptr != NULL)                // Process each line of the config file
  {
      tok = strtok(cptr, " \n");       // Process each string separated by " "
      if (strcmp(tok, "output") == 0)  // Specify output reporting level
      {
         tok = strtok(NULL, " \n");
         if (tok == NULL)
            OutputLevel = 0;
         else OutputLevel = atoi(tok);
      }

      else if (strcmp(tok, "hierarchical") == 0)  // Generate hierarchical SVM
      {
         tok = strtok(NULL, " \n");
         if (tok == NULL)
         {
            printf("hierarchical: config file error\n");
            exit(1);
         }
         hierarchical = atoi(tok);
      }

      else if (strcmp(tok, "tree") == 0)  // File specifying hierarchy structure
      {
         tok = strtok(NULL, " \n");
         treefname = (char *) malloc(sizeof(char) * MAXSTR);
         if (tok == NULL)
            strcpy(treefname, "ar.tree");
         else strcpy(treefname, tok);
      }

      else if (strcmp(tok, "data") == 0)  // Data file name(s)
      {
         tok = strtok(NULL, " \n");
         NumFiles = 0;
         if (tok == NULL)
         {
            printf("No data file specified\n");
            exit(1);
         }
         while (tok != NULL)
         {
            if (NumFiles == 0) filenames = (char **) malloc(sizeof(char *));
            else filenames =
                  (char **) realloc(filenames, sizeof(char *) * (NumFiles + 1));
            filenames[NumFiles] = (char *) malloc(sizeof(char ) * MAXSTR);
            strcpy(filenames[NumFiles], tok);
            NumFiles++;
            tok = strtok(NULL, " \n");
         }
      }

      else if (strcmp(tok, "mode") == 0)   // Train or test mode
      {
         tok = strtok(NULL, " \n");
         Mode = atoi(tok);
      }

      else if (strcmp(tok, "weight") == 0)  // How to weight events in window
      {
         tok = strtok(NULL, " \n");
         Weight = atoi(tok);
      }

      else if (strcmp(tok, "context") == 0) // Include previous activity feature
      {
         tok = strtok(NULL, " \n");
         Context = atoi(tok);
      }

      else if (strcmp(tok, "model") == 0)  // Specify model directory name
      {
         tok = strtok(NULL, " \n");
         if (tok != NULL)
         {
            modelfname = (char *) malloc(sizeof(char) * MAXSTR);
            strcpy(modelfname, tok);
         }
      }

      else if (strcmp(tok, "sensor") == 0)  // List of sensor names
      {
         tok = strtok(NULL, " \n");
         while (tok != NULL)
         {
            if (NumSensors == 0)
               sensornames = (char **) malloc(sizeof(char *));
            else sensornames =
                  (char **) realloc(sensornames, sizeof(char *) * (NumSensors+1));
            sensornames[NumSensors] = (char *) malloc(sizeof(char) * MAXSTR);
            strcpy(sensornames[NumSensors], tok);
            NumSensors++;
            tok = strtok(NULL, " \n");
         }
      }

      else if (strcmp(tok, "activity") == 0)  // List of activity labels
      {
         tok = strtok(NULL, " \n");
         while (tok != NULL)
         {
            if (NumActivities == 0)
               activitynames = (char **) malloc(sizeof(char *));
            else activitynames =
                  (char **) realloc(activitynames,
                          sizeof(char *) * (NumActivities+1));
            activitynames[NumActivities] =
               (char *) malloc(sizeof(char) * MAXSTR);
            strcpy(activitynames[NumActivities], tok);
            NumActivities++;
            tok = strtok(NULL, " \n");
         }
      }

      else if (strcmp(tok, "fsenact") == 0)  // Activity likelihood for sensors
      {
         tok = strtok(NULL, " \n");
         i = 0;
         while (tok != NULL)
         {
            if (i == 0)
               fsenact = (int *) malloc(sizeof(int));
            else fsenact = (int *) realloc(fsenact, sizeof(int) * (i+1));
            fsenact[i] = atoi(tok);
            i++;
            tok = strtok(NULL, " \n");
         }
      }

      else if (strcmp(tok, "factwind") == 0) // Activity probs for window sizes
      {
         tok = strtok(NULL, " \n");
         i = 0;
         while (tok != NULL)
         {
            if (i == 0)
               factwind = (int *) malloc(sizeof(int));
            else factwind = (int *) realloc(factwind, sizeof(int) * (i + 1));
            factwind[i] = atoi(tok);
            i++;
            tok = strtok(NULL, " \n");
         }
      }

      else if (strcmp(tok, "mi") == 0)  // Sensor mutual information
      {
         tok = strtok(NULL, " \n");

         if (tok != NULL)
         {
            MI = (float **) malloc(sizeof(float *) * NumSensors);
            for (i = 0; i < NumSensors; i++)
            {
               MI[i] = (float *) malloc(sizeof(float) * NumSensors);
               for (j = 0; j < NumSensors; j++)
               {
                  MI[i][j] = (float)atof(tok);
                  tok = strtok(NULL, " \n");
               }
            }
         }
      }

      cptr =  fgets(buffer, 40000, fp);  // Read next line
   }

   fclose(fp);
}


/* Print program parameter values */
void PrintParms()
{
   int i;

   printf("*****************************\n");
   printf("*** AR Program Parameters ***\n\n");
   if (Mode == 0)
      printf("Mode: Training mode\n");
   else printf("Mode: Testing mode\n");

   if (Weight == 0)
      printf("Weight: No sensor event weighting\n");
   else if (Weight == 1)
      printf("Weight: Time sensor event weighting\n");
   else if (Weight == 2)
      printf("Weight: Mutual information sensor event weighting\n");
   else printf("Weight: Time and mutual information sensor event weighting\n");

   if (Context == 0)
      printf("Context: No activity context used in model\n");
   else printf("Context: Including previous activity in model\n");

   if (hierarchical == TRUE)
      printf("SVM: using hierarchy\n");
   else printf("SVM: flat model\n");

   if (treefname != NULL)
      printf("tree file: %s\n", treefname);

   if (modelfname != NULL)
      printf("model file: %s\n", modelfname);

   printf("Sensor names:");
   for (i=0; i<NumSensors; i++)
      printf(" %s", sensornames[i]);
   printf("\n");

   if (Mode != TRAIN)
   {
      printf("Activity labels: [%d]", NumActivities);
      for (i=0; i<NumActivities; i++)
         printf(" %s", activitynames[i]);
      printf("\n");
   }
   printf("*****************************\n\n");
}


/* Save the parameters for use during testing */
void SaveParams()
{
   FILE *fp;
   char str[MAXSTR];
   int i, j;

   sprintf(str, "%s/%s", modelfname, "model.config");
   fp = fopen(str, "w");
   if (fp == NULL)
   {
      printf("Unable to open %s for writing\n", str);
      exit(-1);
   }

   // sensor names
   fprintf(fp, "sensor ");
   for (i=0; i<NumSensors; i++)
      fprintf(fp, "%s ", sensornames[i]);
   fprintf(fp, "\n");

   // activity names
   fprintf(fp, "activity ");
   for (i=0; i<NumActivities; i++)
      fprintf(fp, "%s ", activitynames[i]);

   fprintf(fp, "\n");

   // hierarchical
   fprintf(fp, "hierarchical %d\n", hierarchical);

   // tree structure
   fprintf(fp, "tree %s\n", treefname);

   // fsenact
   fprintf(fp, "fsenact ");
   for (i=0; i<NumSensors; i++)
      fprintf(fp, "%d ", fsenact[i]);
   fprintf(fp, "\n");

   // factwind
   fprintf(fp, "factwind ");
   for (i=0; i<NumActivities-TreeDepth; i++)
      fprintf(fp, "%d ", factwind[i]);
   fprintf(fp, "\n");

   // weight
   fprintf(fp, "weight %d\n", Weight);

   // context
   fprintf(fp, "context %d\n", Context);

   // mi
   fprintf(fp, "mi ");
   for (i=0; i<NumSensors; i++)
      for (j=0; j<NumSensors; j++)
         fprintf(fp, "%.6f ", MI[i][j]);
   fprintf(fp, "\n");

   // model directory
   fprintf(fp, "model %s\n", modelfname);

   // test mode
   fprintf(fp, "mode 1\n");

   fclose(fp);

   if (OutputLevel > 0)
      printf("Parameters saved to file.\n");
}


/* Find the index of a specified activity from the activity list */
int FindActivity(char *name)
{
   int i;

   for (i=0; i<NumActivities; i++)
      if (strcmp(name, activitynames[i]) == 0)
         return i;

   if (NumActivities == 0)
   {
      activitynames = (char **) malloc(sizeof(char *));
      actwind = (int **) malloc(sizeof(int *));
   }
   else
   {
      activitynames =
          (char **) realloc(activitynames, sizeof(char *) * (NumActivities+1));
      actwind = (int **) realloc(actwind, sizeof(int *) * (NumActivities+1));
   }	
   activitynames[NumActivities] = (char *) malloc(sizeof(char) * MAXSTR);
   strcpy(activitynames[NumActivities], name);

   actwind[NumActivities] = (int *) calloc(NUMWINDOWS, sizeof(int));
   NumActivities++;

   return(NumActivities-1);
}


/* Find the index of a specified sensor from the sensor list */
int FindSensor(char *name)
{
    int i;

    for (i=0; i<NumSensors; i++)
        if (strcmp(name, sensornames[i]) == 0)
            return(i);
    printf("could not find the sensor %s\n", name);
    exit(-1);
}


/* Compute the probability distribution over activity labels for alternative
   window sizes.  Determine the most likely window size given an activity.    */
void ComputeProbActWind()
{
   FILE *fp;
   char *cptr, buffer[256], sdate[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR] , sensorstatus[MAXSTR], alabel[MAXSTR];
   char prevlabel[MAXSTR];
   int cnt, i, j, anum, max, numfile;

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s\n", filenames[numfile]);
         exit(1);
      }

      cptr = fgets(buffer, 256, fp);
      sscanf(buffer, "%s %s %s %s %s %s",
               sdate, stime, sensorid, newsensorid, sensorstatus, alabel);

      strcpy(prevlabel, alabel);  // Keep track of previous activity
      cnt = 0;
      while (cptr != NULL)  // Process each sensor event in the file
      {
          sscanf(buffer, "%s %s %s %s %s %s",
                  sdate, stime, sensorid, newsensorid, sensorstatus, alabel);

          // Store activity probability distribution for
          // four alternative window sizes: 5, 10, 20, and 30
          if (strcmp(prevlabel, alabel) != 0)
          {
             anum = FindActivity(prevlabel);
             if (cnt <= 5)
                actwind[anum][0]++;
             else if (cnt > 5 && cnt <= 10)
                actwind[anum][1]++;
             else if (cnt > 10 && cnt <= 20)
                actwind[anum][2]++;
             else if (cnt > 20)
                actwind[anum][3]++;
             cnt = 0;
             strcpy(prevlabel, alabel);
          }
          else cnt++;

          cptr = fgets(buffer, 256, fp);
       }
      fclose(fp);
   }

   // Identify the window size with the greatest probability for each activity
   factwind = (int *) malloc(sizeof(int) * NumActivities);
   for (i=0; i<NumActivities; i++)
   {
      max = 0;
      for (j=0; j<NUMWINDOWS; j++)
      {
         if (actwind[i][j] > max)
         {
            factwind[i] = j;
            max = actwind[i][j];
         }
      }
   }
}


/* Compute the probability distribution of the activities given the sensor and
   identify the most likely activity given the sensor */
void ComputeProbSenAct()
{
   FILE *fp;
   char *cptr, buffer[256], sdate[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   int i, j, snum, anum, max, numfile;

   senact = (int **) malloc(sizeof(int *) * NumSensors);
   for (i=0; i<NumSensors; i++)
      senact[i] = (int *) calloc(NumActivities, sizeof(int));

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s\n", filenames[numfile]);
         exit(1);
      }

      cptr = fgets(buffer, 256, fp);
      while (cptr != NULL)  // Process each sensor event in the file
      {
         sscanf(buffer, "%s %s %s %s %s %s",
                   sdate, stime, sensorid, newsensorid, sensorstatus, alabel);
         if (strcmp(alabel, "none") != 0)
         {
            snum = FindSensor(newsensorid);
            anum = FindActivity(alabel);
            senact[snum][anum]++;
         }
         cptr = fgets(buffer, 256, fp);
      }
      fclose(fp);
   }

   // Identify most likely activity associated with each sensor
   fsenact = (int *) calloc(NumSensors, sizeof(int));
   for (i=0; i<NumSensors; i++)
   {
      max = 0;
      for (j=0; j<NumActivities; j++)
      {
         if (senact[i][j] > max)
         {
            fsenact[i] = j;
            max = senact[i][j];
         }
      }
   }
}


/* Compute the mutual information between sensors by determining the
   probability of co-occurence of every pair of sensors */
void ComputeMI()
{
   FILE *fp;
   char *cptr, buffer[256], sdate[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   int i, j, numfile, presen, cursen;
   float total;

   MI = (float **) malloc(sizeof(float *) * NumSensors);
   for (i=0; i<NumSensors; i++)
      MI[i] = (float *) calloc(NumSensors, sizeof(float));

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s\n", filenames[numfile]);
         exit(1);
      }

      presen = 0;
      cursen = 0;

      cptr = fgets(buffer, 256, fp);
      while (cptr != NULL)  // Process each sensor event in the file
      {
         sscanf(buffer, "%s %s %s %s %s %s",
                   sdate, stime, sensorid, newsensorid, sensorstatus, alabel);
         cursen = FindSensor(newsensorid);

         // Update the MI count based on the temporal proximity of presen
         // (the previous sensor) to cursen (the current sensor)
         MI[presen][cursen]++;
         presen = cursen;
         cptr = fgets(buffer, 256, fp);
      }
      fclose(fp);
   }

   for (i=0; i<NumSensors; i++)  // Normalize MI values for each pair of sensors
   {
      total = 0;
      for (j=0; j<NumSensors; j++)
         total += MI[i][j];
      if (total != 0)
      {
         for (j=0; j<NumSensors; j++)
            MI[i][j] /= total;
      }
   }
}


/* Read the data from the input files.  Compute a feature vector for
   each window-length sequence of sensor events. */
void ReadData()
{
   FILE *fp;
   char *cptr, buffer[256], date[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   int i, cnt, numfile, temp;
   float *tempdata;

   NumFeatures = NumSensors + 12; // Feature length is #sensors + 12 time values

   numdataperactivity = (int *) calloc(NumActivities, sizeof(int));
   adindices = (int **) malloc(sizeof(int *) * NumActivities);
   tempdata = (float *) malloc(sizeof(float) * NumFeatures);

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      if (Mode != TEST)
         printf("Reading file %s ", filenames[numfile]);
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s", filenames[numfile]);
         exit(1);
      }
      cptr = fgets(buffer, 256, fp);

      cnt = 0;	
      while (cptr != NULL)  // Process each sensor event in the file
      {
         sscanf(buffer, "%s %s %s %s %s %s",
                  date, stime, sensorid, newsensorid, sensorstatus, alabel);
         // initialize the temp data
         for (i=0; i<NumFeatures; i++)
            tempdata[i] = 0;

         temp = ComputeFeature(stime, newsensorid, tempdata);
         if (temp == 1)  // End of window reached, add feature vector
         {
            if (NumData == 0)
               data = (float **) malloc(sizeof(float *));
            else data =
	       (float **) realloc(data, sizeof(float *) * (NumData+1));
            data[NumData] = (float *) malloc(sizeof(float) * NumFeatures);

            for (i=0; i<NumFeatures; i++)
               data[NumData][i] = tempdata[i];

            AddDataToActivity(alabel, NumData);
            NumData++;
         }
         cptr = fgets(buffer, 256, fp);
      }
      fclose(fp);
   }

   free(tempdata); // free variables used by this function
}


/* Compute the feature vector for each window-size sequence of sensor events.
   The feature vector consists of:
   fields  1 -  4: quantized value for window start time
   5 -  8: quantized value for window end time
   9 - 12: quantized value for window time duration
   13+:     count of events for each unique sensor                 */
int ComputeFeature(char *time, char *sensorid, float *tempdata)
{
   int i, wsize, duration;

   windata[wincnt][0] = ComputeHour(time);
   windata[wincnt][1] = FindSensor(sensorid);
   windata[wincnt][2] = ComputeSeconds(time);

   if (wincnt < (MAXWINDOW-1))  // Not reached end of window
   {
      wincnt++;  // Keep track of number of events in current window
      return(0);
   }
   else // Reached end of window, create feature vector
   {
      wsize = GetWindow(sensorid); // determine the window size

      // Sensor attributes
      // Weight each sensor event within the window
      // according to the selected weight scheme (none, time, MI, or time+MI)
      for (i=(MAXWINDOW-1); i>(MAXWINDOW-wsize); i--)
      {
         if (Weight == NONE)
            tempdata[windata[i][1]+12]++;

         else if (Weight == TIMEWEIGHT)
            tempdata[windata[i][1]+12] += (float) exp(-KAPPA *
                         fabs(windata[MAXWINDOW-1][2] - windata[i][2]));

         else if (Weight == MUTUALINFO)
            tempdata[windata[i][1]+12] +=
                 MI[windata[MAXWINDOW-1][1]][windata[wincnt][1]];

         else if (Weight == MUTUALINFO_TIME)
            tempdata[windata[i][1]+12] += (float) exp(-KAPPA *
                         fabs(windata[MAXWINDOW-1][2] - windata[i][2])) *
                         MI[windata[MAXWINDOW-1][1]][windata[wincnt][1]];
      }

      // time attributes
      tempdata[windata[MAXWINDOW-wsize][0]/6]++;      // start time
      tempdata[windata[MAXWINDOW-1][0]/6+4]++;        // end time
      // time duration of window
      duration = abs(windata[MAXWINDOW-wsize][0] - windata[MAXWINDOW-1][0]);
      if (duration == 0)
          tempdata[8]++;
      else if (duration == 1)
          tempdata[9]++;
      else if (duration <=3)
          tempdata[10]++;
      else tempdata[11]++;

      for (i=0; i<MAXWINDOW-1; i++)
      {
         windata[i][0] = windata[i+1][0];
         windata[i][1] = windata[i+1][1];
         windata[i][2] = windata[i+1][2];
      }
      return(1);
   }

   return(0);
}


/* Compute the hour of the day (based on 24 hour clock). */
int ComputeHour(char *time)
{
   char temp[3];

   temp[0] = time[0];
   temp[1] = time[1];
   temp[2] = '\0';

   return(atoi(temp));
}


/* Compute the number of seconds past midnight on the current day. */
int ComputeSeconds(char *time)
{
   char hour[3], minute[3], second[3];

   hour[0] = time[0];
   hour[1] = time[1];
   hour[2] = '\0';
   minute[0] = time[3];
   minute[1] = time[4];
   minute[2] = '\0';
   second[0] = time[6];
   second[1] = time[7];
   second[2] = '\0';

   return((atoi(hour)*3600)+(atoi(minute)*60)+atoi(second));
}


/* Compute the window size for a given sensor. */
int GetWindow(char *name)
{
   int win;

   win = factwind[fsenact[FindSensor(name)]];

   if (win == 0)
      return 5;
   else if (win == 1)
      return 10;
   else if (win == 2)
      return 20;
   else return 30;
}


/* Add the data index to the activity list. */
void AddDataToActivity(char *name, int count)
{
   int i;

   for (i=0; i<NumActivities; i++)  // Create an entry for each activity
      if (strcmp(name, activitynames[i]) == 0)  // Add data to activity list
      {
         if (numdataperactivity[i] > 0)
         {
            adindices[i] = (int *) realloc(adindices[i],
                      sizeof(int) * (numdataperactivity[i]+1));
            adindices[i][numdataperactivity[i]] = count;
            numdataperactivity[i]++;
            return;
         }
         else
         {
            adindices[i] = (int *) malloc(sizeof(int));
            adindices[i][numdataperactivity[i]] = count;
            numdataperactivity[i]++;
            return;
         }
      }

   if (NumActivities == 0)  // Called for the first time, initialize variables
   {
      activitynames = (char **) malloc(sizeof(char *));
      adindices = (int **) malloc(sizeof(int *));
      numdataperactivity = (int *) malloc(sizeof(int));
   }
   else 
   {
      activitynames =
           (char **) realloc(activitynames, sizeof(char *) * (NumActivities+1));
      adindices =
           (int **) realloc(adindices, sizeof(int *) * (NumActivities+1));
      numdataperactivity = (int *) realloc(numdataperactivity,
               sizeof(int) * (NumActivities+1));
   }
   activitynames[NumActivities] = (char *) malloc(sizeof(char) * MAXSTR);
   adindices[NumActivities] = (int *) malloc(sizeof(int));
   numdataperactivity[NumActivities] = 0;

   strcpy(activitynames[NumActivities], name);
   adindices[NumActivities][numdataperactivity[NumActivities]] = count;
   numdataperactivity[NumActivities]++;
   NumActivities++;
}


/* Read the tree structure to use for a hierarchical SVM. */
void ReadTree()
{
   FILE *fp;
   char *cptr, buffer[512], *tok;

   fp = fopen(treefname, "r");
   if (fp == NULL)
   {
      printf("Unable to open tree file: %s\n", treefname);
      exit(-1);
   }
   cptr =  fgets(buffer, 512, fp);

   TreeDepth = 0;

   // Process each line of the file.
   // Each line contains a node name and a list of the node's children.
   // Children can be activity labels or names of other nodes.
   while (cptr != NULL)
   {
      tok = strtok(cptr, " \n");
      if (tok != NULL)
      {
          if (TreeDepth == 0)
          {
              tree = (char ***) malloc(sizeof(char **));
              NumSiblings = (int*) malloc(sizeof(int));
          }
          else
          {
              tree = (char ***) realloc(tree, sizeof(char **) * (TreeDepth+1));
              NumSiblings = (int *) realloc(NumSiblings,
                      sizeof(int) * (TreeDepth+1));
          }
          NumSiblings[TreeDepth] = 0;

          while (tok != NULL)
          {
              if (NumSiblings[TreeDepth] == 0)
                  tree[TreeDepth] = (char **) malloc(sizeof(char *));
              else tree[TreeDepth] =
                  (char **) realloc(tree[TreeDepth],
                          sizeof(char *) * (NumSiblings[TreeDepth]+1));

              tree[TreeDepth][NumSiblings[TreeDepth]] =
                  (char *) malloc(sizeof(char) * MAXSTR);
              strcpy(tree[TreeDepth][NumSiblings[TreeDepth]], tok);
              NumSiblings[TreeDepth]++;
              tok = strtok(NULL, " \n");
          }
          TreeDepth++;
      }
      cptr =  fgets(buffer, 512, fp);
   }
   fclose(fp);
}


/* Build a default tree (flat model).
   The flat model has one root whose children are all of the activity labels. */
void DefaultTree()
{
   int i;

   TreeDepth = 0;
   tree = (char ***) malloc(sizeof(char **));
   NumSiblings = (int*) malloc(sizeof(int));
   NumSiblings[TreeDepth] = NumActivities+1;
   tree[TreeDepth] = (char **) malloc(sizeof(char *) * NumSiblings[TreeDepth]);
   tree[TreeDepth][0] = (char *) malloc(sizeof(char) * MAXSTR);

   strcpy(tree[TreeDepth][0], "node_all");

   for (i=1; i<NumSiblings[TreeDepth]; i++)
   {
      tree[TreeDepth][i] = (char *) malloc(sizeof(char) * MAXSTR);
      strcpy(tree[TreeDepth][i], activitynames[i-1]);
   }
   TreeDepth++;
}


/* Create a SVM model for every node in the hierarchy. */
void LearnActivityModels()
{
   char str[MAXSTR];
   int i, j, depth, sibling, actindex;
   int *labels, numlabel, probability = 0, derivecontext, cval, numcontext;
   float *datacontext;

   if (Context == 1)
      probability = 1;

   for (depth=0; depth<TreeDepth; depth++)
   {
      numlabel = 0;
      for (sibling=1; sibling<NumSiblings[depth]; sibling++)
      {
         for (actindex=0; actindex<NumActivities; actindex++)
            if (strcmp(activitynames[actindex], tree[depth][sibling]) == 0)
            {
               if (numlabel == 0)
                  labels = (int *) malloc(sizeof(int));
               else
                  labels = (int *) realloc(labels, sizeof(int) * (numlabel+1));
               labels[numlabel] = actindex;
               numlabel++;
               break;
            }
      }

      // train the model for the current node
      if (Context == TRUE)
      {
         if (OutputLevel > 0)
            printf("Training the context model for %s...", tree[depth][0]);
         sprintf(str, "%s/%s_context", modelfname, tree[depth][0]);
         TrainSVM(labels, numlabel, str, probability);
      }
      else
      {
         if (OutputLevel > 0)
            printf("Training the model for %s...", tree[depth][0]);
         sprintf(str, "%s/%s", modelfname, tree[depth][0]);
         TrainSVM(labels, numlabel, str, probability);
      }
      if (OutputLevel > 0)
         printf("done\n");

      // store the indices for the new "activity" except for the root node
      if (depth < TreeDepth-1)
      {
         for (sibling=1; sibling < NumSiblings[depth]; sibling++)
         {
            for (actindex=0; actindex<NumActivities; actindex++)
               if (strcmp(activitynames[actindex], tree[depth][sibling]) == 0)
               {
                  for (i=0; i<numdataperactivity[actindex]; i++)
                     AddDataToActivity(tree[depth][0], adindices[actindex][i]);
                  break;
               }
         }
      }

      if (numlabel > 0)
      {
         free(labels);
         numlabel = 0;
      }
   }

   if (Context == TRUE)  // Keep track of previous activity and include in model
   {
      LoadModels(1); // load the context models

      // compute probability distributions for data with the previous activity
      // keep adding the probability values to the data
      numcontext = 0;
      for (depth=0; depth<TreeDepth; depth++)
          numcontext += NumSiblings[depth] - 1;

      numcontext *= 2;

      datacontext = (float *) calloc(numcontext, sizeof(float));
      derivecontext = TRUE;

      for (i=0; i<NumData; i++)
      {
         cval = TestSVM(derivecontext, data[i], datacontext, numcontext);            
         // add the probability distributions and the previous activity context
         // to the feature vector
         data[i] = (float *) realloc(data[i],
                  sizeof(float) * (NumFeatures+numcontext));
         for (j=0; j<numcontext; j++)
         {
            data[i][j+NumFeatures] = datacontext[j];
            datacontext[j] = 0;
         }
      }
      free(datacontext);
      NumFeatures += numcontext;

      probability = 0;

      // train the models for the second level
      for (depth=0; depth<TreeDepth; depth++)
      {
         numlabel = 0;
         for (sibling=1; sibling<NumSiblings[depth]; sibling++)
         {
            for (actindex=0; actindex<NumActivities; actindex++)
               if (strcmp(activitynames[actindex], tree[depth][sibling]) == 0)
               {
                  if (numlabel == 0)
                     labels = (int *) malloc(sizeof(int));
                  else labels = (int *) realloc(labels,
                         sizeof(int) * (numlabel+1));
                  labels[numlabel] = actindex;
                  numlabel++;
                  break;
               }
         }

         // train the model for the current node
         if (OutputLevel > 0)
            printf("Training the model for %s...", tree[depth][0]);
         sprintf(str, "%s/%s", modelfname, tree[depth][0]);
         TrainSVM(labels, numlabel, str, probability);
         if (OutputLevel > 0)
            printf("done\n");

         if (numlabel > 0)
         {
            numlabel = 0;
            free(labels);
         }
      }
   }
}


/* Train an SVM on an input set of feature values. */
void TrainSVM(int *labels, int numlabel, char *modelfilename, int probability)
{
   int i, j, cnt, sibling;
   int actindex, idxcnt;
   struct svm_problem *prob;
   struct svm_parameter parameter;
   struct svm_model *model;

   // parameters for the SVM model
   parameter.svm_type = C_SVC;         // Multi-Class SVM (default is pairwise)
   parameter.kernel_type = RBF;        // Kernel type
   parameter.degree = 3;
   parameter.gamma = GAMMA;		// Gaussian width for RBF kernel
   parameter.coef0 = 0;
   parameter.nu = 0.5;
   parameter.cache_size = 100;
   parameter.C = C;			// Penalty parameter
   parameter.eps = 1e-3;
   parameter.p = 0.1;
   parameter.shrinking = 0;
   parameter.probability = probability;
   parameter.nr_weight = 0;
   parameter.weight_label = NULL;

   prob = (struct svm_problem *) malloc(sizeof(struct svm_problem));
   prob->l = 0;
   for (sibling=0; sibling<numlabel; sibling++)
   {
      actindex = labels[sibling];
      for (i=0; i<numdataperactivity[actindex]; i++)
      {
         if (prob->l == 0)
         {
            prob->x = (struct svm_node **) malloc(sizeof(struct svm_node *));
            prob->y = (double *) malloc(sizeof(double));
         }
         else
         {
            prob->x = (struct svm_node **) realloc(prob->x,
                     sizeof(struct svm_node *) * (prob->l+1));
            prob->y = (double *) realloc(prob->y, sizeof(double) * (prob->l+1));
         }

         idxcnt = prob->l;
         prob->x[idxcnt] =
             (struct svm_node *) malloc(sizeof(struct svm_node) * NumFeatures);

         cnt = 0;
         for (j=0; j<NumFeatures; j++)
         {
            if (data[adindices[actindex][i]][j] != 0)
            {
               prob->x[idxcnt][cnt].index = j;
               prob->x[idxcnt][cnt].value = data[adindices[actindex][i]][j];
               cnt++;
            }
         }
         prob->x[idxcnt][cnt].index = -1;
         prob->x[idxcnt][cnt].value = 0;
         prob->y[idxcnt] = actindex;
         prob->l++;
      }
   }
   // train the model for the current node
   model = svm_train(prob, &parameter);

   // save the model for the current node using the provided name
   svm_save_model(modelfilename, model);

   // free the current problem structure and model for use on the next iteration
   svm_free_model_content(model);
   for (i=0; i<prob->l; i++)
      free(prob->x[i]);
   if (prob->l > 0)
   {
      free(prob->x);
      free(prob->y);
      prob->l = 0;
   }
   free(prob);
}


/* Test the SVM model for every node in the hierarchy. */
void TestActivityModels()
{
   int i, j, depth, derivecontext, cval, numcontext = 0, temp=0;
   float *datacontext;

   if (Context == TRUE) // load the context models
      LoadModels(1);
   LoadModels(0);    // load the regular models

   for (depth=0; depth<TreeDepth; depth++)
      for (j=0; j<NumSiblings[depth]; j++)
      {
         temp ++;
         if (strcmp(tree[depth][j], "Other_Activity") == 0)
            OtherActivityIndex = temp;
      }

   if (Context == TRUE)
   {
      // compute probability distribution for the data and the previous activity
      // keep adding the probability values to the data
      numcontext = 0;
      for (depth=0; depth<TreeDepth; depth++)
         numcontext += NumSiblings[depth] - 1;

      numcontext *= 2;

      datacontext = (float *) calloc(numcontext, sizeof(float));
      derivecontext = TRUE;
   }

   for (i=0; i<NumData; i++)
   {
      if (Context == TRUE)
      {
          cval = TestSVM(derivecontext, data[i], datacontext, numcontext);

          // add the probability distributions and previous activity context
          // to the feature vector
          data[i] = (float * ) realloc(data[i],
                   sizeof(float) * (NumFeatures+numcontext));
          for (j=0; j<numcontext; j++)
          {
             data[i][j+NumFeatures] = datacontext[j];
             datacontext[j] = 0;
          }
      }

      cval = TestSVM(0, data[i], datacontext, numcontext);
      printf("%s\n", activitynames[cval]);
   }

   if (Context == TRUE)
   {
      NumFeatures += numcontext;
      free(datacontext);
   }
}


/* Test the SVM model on input data. */
int TestSVM(int derivecontext, float *tempdata,
        float *datacontext, int numcontext)
{
   char *strptr;
   int j, cnt, depth, nfeature;
   int tempnumfeatures, maxindex;
   float maxval;
   double *prob_values, cval;
   struct svm_node *x;

   if (derivecontext == TRUE)
      nfeature = NumFeatures;
   else nfeature = NumFeatures + numcontext;

   // convert the data array into the svm format
   x = (struct svm_node*) malloc(sizeof(struct svm_node) * nfeature);
   cnt = 0;
   for (j=0; j<nfeature; j++)
   {
      if (tempdata[j] != 0)
      {
         x[cnt].index = j;
         x[cnt].value = tempdata[j];
         cnt++;
      }
   }
   x[cnt].index = -1;
   x[cnt].value = 0;

   // if derive probabilities
   if (derivecontext ==  TRUE)
   {
      // find the class probability distributions
      tempnumfeatures = 0;
      for (depth=0; depth<TreeDepth; depth++)
      {
         prob_values =
              (double *) malloc(sizeof(double) * (NumSiblings[depth] -1));
         cval = svm_predict_probability(contextmodels[depth], x, prob_values);
         for (j=0; j<(NumSiblings[depth]-1); j++)
            datacontext[tempnumfeatures+j] = (float) prob_values[j];
         tempnumfeatures += NumSiblings[depth] - 1;
         free(prob_values);
      }

      maxval = 0;
      maxindex = 0;

      // predict the activity and compare it with the previous activity
      for (j=0; j<tempnumfeatures; j++)
         if (maxval < datacontext[j])
         {
            maxval = datacontext[j];
            maxindex = j;
         }

      if (maxindex == CurActivity)
         datacontext[PrevActivity+tempnumfeatures] = 1;
      else
      {
         if (CurActivity != OtherActivityIndex) // last index is Other_Activity
         {
            datacontext[CurActivity+tempnumfeatures] = 1;
            PrevActivity = CurActivity;
            CurActivity = maxindex;
         }
         else datacontext[PrevActivity+tempnumfeatures] = 1;
      }

      free(x);
      return(1);
   }
   else
   {
      cval = svm_predict(models[TreeDepth-1], x);

      strptr = strstr(activitynames[(int)cval], "node");
      if (strptr != NULL)
      {
         for (depth=0; depth<TreeDepth-1; depth++)
            if (strcmp(tree[depth][0], activitynames[(int)cval]) == 0)
            {
               cval = svm_predict(models[depth], x);
               break;
            }
      }

      free(x);
      return((int) cval);
   }
}


/* Read the translate file.  This file contains a mapping from sensor
   identifiers as they appear in the input data file and the corresponding
   sensor identifiers as they will be used in the model.  There are three
   strings on each line:  the original identifier and two sensor names.
   The convention is that the first sensor name is more abstract (usually
   a room name) and the second is more specific (an area within the room
   where the sensor is located).                                              */
void ReadTranslate()
{
   FILE *fp;
   char *cptr, buffer[256], sensorid[MAXSTR], translatesensorid[MAXSTR];

   fp = fopen(translatefilename, "r");
   if (fp == NULL)
   {
      printf("Unable to open the translate file name: %s", translatefilename);
      exit(-1);
   }

   NumTranslate = 0;
   cptr = fgets(buffer, 256, fp);
   while (cptr != NULL)  // Process each line in the file
   {
      sscanf(cptr, "%s %s", sensorid, translatesensorid);
      if (NumTranslate == 0)
         translate = (char ***) malloc(sizeof(char **));
      else translate = (char ***) realloc(translate,
             sizeof(char **) * (NumTranslate + 1));
      translate[NumTranslate] = (char**) malloc(sizeof(char *) * 2);
      translate[NumTranslate][0] = (char *) malloc(sizeof(char) * MAXSTR);
      translate[NumTranslate][1] = (char *) malloc(sizeof(char) * MAXSTR);
      strcpy(translate[NumTranslate][0], sensorid);
      strcpy(translate[NumTranslate][1], translatesensorid);
      cptr = fgets(buffer, 256, fp);
   }
   fclose(fp);
}


/* Read the activity models, one model for each node in the hierarchy. */
void LoadModels(int context)
{
   char str[MAXSTR];
   int depth = 0;

   if (context == TRUE)
   {
      contextmodels =
            (struct svm_model **) malloc(sizeof(struct svm_model *) * TreeDepth);
      for (depth=0; depth<TreeDepth; depth++)
      {
         sprintf(str, "%s/%s_context", modelfname, tree[depth][0]);
         contextmodels[depth] = svm_load_model(str);
      }
   }
   else
   {
      models =
         (struct svm_model **) malloc(sizeof(struct svm_model *) * TreeDepth);
      for (depth=0; depth<TreeDepth; depth++)
      {
         sprintf(str, "%s/%s", modelfname, tree[depth][0]); 
         models[depth] = svm_load_model(str);
         if (models[depth] == NULL)
         {
	    printf("error in reading the model file %s\n", str);
	    exit(1);
	 }
      }
   }
}


/* Free allocated memory. */
void FreeMemory()
{
   int i, j;

   if (NumFiles > 0)
   {
      for (i=0; i<NumFiles; i++)
         free(filenames[i]);
      free(filenames);
   }
   if (modelfname != NULL)
      free(modelfname);
   if (treefname != NULL)
      free(treefname);
   if (NumSensors > 0)
   {
      for (i=0; i<NumSensors; i++)
         free(sensornames[i]);
      free(sensornames);
   }
   if (senact != NULL)
   {
      for (i=0; i<NumSensors; i++)
         free(senact[i]);
      free(senact);
   }
   if (actwind != NULL)
   {
      for (i=0; i<NumActivities-TreeDepth; i++)
         free(actwind[i]);
      free(actwind);
   }
   if (factwind != NULL)
      free(factwind);
   if (fsenact != NULL)
      free(fsenact);
   if (MI != NULL)
   {
      for (i=0; i<NumSensors; i++)
         free(MI[i]);
      free(MI);
   }
   if (NumData > 0)
   {
      for (i=0; i<NumData; i++)
         free(data[i]);
      free(data);
   }
   if (NumActivities > 0)
   {
      for (i=0; i<NumActivities; i++)
      {
        if (numdataperactivity[i] > 0)
           free(adindices[i]);
        free(activitynames[i]);
      }
      free(adindices);
      free(activitynames);
      free(numdataperactivity);
   }
   if (TreeDepth > 0)
   {
      for(i = 0; i < TreeDepth; i++)
      {
         for (j=0; j<NumSiblings[i]; j++)
            free(tree[i][j]);
         free(tree[i]);
      }
      free(tree);
      free(NumSiblings);
   }
   if (contextmodels != NULL)
   {
      for (i=0; i<TreeDepth; i++)
         svm_free_model_content(contextmodels[i]);
      free(contextmodels);
   }
   if (models != NULL)
   {
      for (i=0; i<TreeDepth; i++)
         svm_free_model_content(models[i]);
      free(models);
   }
}
